﻿
#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Optimization;
using HEAL.Attic;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  /// <summary>
  /// An operator that tracks the min average and max length of symbolic expression trees.
  /// </summary>
  [Item("MinAverageMaxSymbolicExpressionNeastedTreeSizeAnalyzer", "An operator that tracks the min avgerage and max Neasted Tree Size of symbolic expression trees.")]
  [StorableType("73ED3D0E-47F8-455D-B8C5-1220ACFEC5E1")]
  public sealed class MinAverageMaxSymbolicExpressionNeastedTreeSizeAnalyzer : SymbolicDataAnalysisAnalyzer, ISymbolicExpressionTreeAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string MinMaxAvgNeastedTreeSizeResultName = "MinMaxAvgNeastedTreeSize";

    #region parameter properties
    public ValueLookupParameter<VariableCollection> ResultsParameter {
      get { return (ValueLookupParameter<VariableCollection>)Parameters[ResultsParameterName]; }
    }
    #endregion

    [StorableConstructor]
    private MinAverageMaxSymbolicExpressionNeastedTreeSizeAnalyzer(StorableConstructorFlag deserializing) : base() { }
    private MinAverageMaxSymbolicExpressionNeastedTreeSizeAnalyzer(MinAverageMaxSymbolicExpressionNeastedTreeSizeAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      AfterDeserialization();
    }
    public MinAverageMaxSymbolicExpressionNeastedTreeSizeAnalyzer()
      : base() {
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MinAverageMaxSymbolicExpressionNeastedTreeSizeAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var trees = SymbolicExpressionTreeParameter.ActualValue;
      //var variablesNumber = new DoubleArray(trees.Length);
      var variablesNumber = trees.Select(tree => tree.IterateNodesPostfix().Sum(n => n.GetLength())).ToArray();
      var min = variablesNumber.Min();
      var max = variablesNumber.Max();
      var avg = variablesNumber.Average();

      DataTable table;
      if (ResultCollection.ContainsKey(MinMaxAvgNeastedTreeSizeResultName)) {
        table = (DataTable)ResultCollection[MinMaxAvgNeastedTreeSizeResultName].Value;
      } else {
        table = new DataTable("Tree Variables Number");
        ResultCollection.Add(new Result(MinMaxAvgNeastedTreeSizeResultName, table));
        table.Rows.Add(new DataRow("Min") { VisualProperties = { StartIndexZero = true } });
        table.Rows.Add(new DataRow("Max") { VisualProperties = { StartIndexZero = true } });
        table.Rows.Add(new DataRow("Avg") { VisualProperties = { StartIndexZero = true } });
      }
      table.Rows["Min"].Values.Add(min);
      table.Rows["Max"].Values.Add(max);
      table.Rows["Avg"].Values.Add(avg);

      return base.Apply();
    }
  }
}
